/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Configuration module.
 * 
 * This file contains the configuration structure internal to waaad daemon.
 * Note that the waaad daemon has no support for reloading the configuration -- at least yet.
 * 
 */

#ifndef _CONF_H
#define _CONF_H

#include <waaad/conf-api.h>
#include <waaad/peer-api.h>

/* This type defines the configuration information */
typedef struct {
	int		 log_to;		/* combination of LOG_TO_* flags from log.h */

	char		*filepath;		/* Location of the configuration file to parse */
	
	ext_conf_t 	 pub;			/* The parts of config that are seen by extensions as well */
	
	size_t		 diamid_len;		/* the length of this peer's diameter id string */
	size_t		 realm_len;		/* the length of this peer's realm string */
	
	size_t		 supported_apps_nb;	/* number of elements in the following array */
	peer_appl_t 	*supported_apps_list;	/* Data about the supported applications; will be filled by the dispatch module. the array is ordered by appid */
	
} waaad_conf_t;

/* Pointer to the global configuration object */
extern waaad_conf_t * g_conf;
#define g_pconf	(&g_conf->pub)

/*
 * FUNCTION:	conf_init
 *
 * PARAMETERS:
 *  - 	
 *
 * DESCRIPTION: 
 *  Initialize the configuration module. Do not access g_conf before this function is called.
 *
 * RETURN VALUE:
 *   0	: Default configuration is now loaded.
 *  !0  : An error occurred. 
 */
int conf_init ( void );

/*
 * FUNCTION:	conf_parse
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *  Parse the configuration file (using the yacc parser) found in g_conf->filepath
 *  Note that this has impact not only on the g_conf variable, but also
 *  on other modules such as extensions, peers or routing. Therefore, this function must
 *  be called only when all modules are initialized.
 *
 * RETURN VALUE:
 *   0	: The configuration file has been parsed without error.
 *  !0  : An error occurred.
 */
int conf_parse ( void );

/*
 * FUNCTION:	conf_fini
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *  Terminates the configuration module.
 *
 * RETURN VALUE:
 *   0 : log facility closed properly.
 *  !0 : an error occurred (we may ignore it)
 */
int conf_fini ( void );

#endif /* _CONF_H */
