/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

#include "tests.h"

/* Test for the dispatch module */

/* vendor id */
#define T_VID	735671
/* app ids */
#define T_AID1	735674	/* Vendor, auth + acct */
#define T_AID2	735677	/* auth only */
#define T_AID3	735675	/* auth + acct */
/* command ids */
#define T_CID1	735679
#define T_CID2	735680
/* AVP ids */
#define T_AVID1	735681
#define T_AVID2	735682
/* Enum val Ids */
#define T_EV1	735688
#define T_EV2	735689

static pthread_mutex_t test_mtx = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  test_cnd = PTHREAD_COND_INITIALIZER;
static uint32_t cb_called = 0;

static dict_object_t * cmd1r = NULL;
static dict_object_t * cmd1a = NULL;


#define DEF_CB_DISP( NR, RESETMSG )					\
static int disp_cb ## NR ( msg_t ** msg, msg_avp_t * avp )		\
{									\
	CHECK( 0, pthread_mutex_lock(&test_mtx) );			\
	cb_called |= 1<< NR ;						\
	CHECK( 0, pthread_cond_signal(&test_cnd) );			\
	CHECK( 0, pthread_mutex_unlock(&test_mtx) );			\
	if (RESETMSG)							\
		*msg = NULL;						\
	return 0;							\
}
/* The callbacks to be called by the dispatch thread */
DEF_CB_DISP( 0, 0 );
DEF_CB_DISP( 1, 0 );
DEF_CB_DISP( 2, 0 );
DEF_CB_DISP( 3, 0 );
DEF_CB_DISP( 4, 0 );
DEF_CB_DISP( 5, 0 );
DEF_CB_DISP( 6, 0 );
DEF_CB_DISP( 7, 0 );
DEF_CB_DISP( 8, 0 );
DEF_CB_DISP( 9, 0 );
DEF_CB_DISP(10, 1 );

static void cbans(void * data, msg_t ** answer)
{
	CHECK( 0, pthread_mutex_lock(&test_mtx) );
	cb_called |= 1<< 31 ;
	CHECK( 0, pthread_cond_signal(&test_cnd) );
	CHECK( 0, pthread_mutex_unlock(&test_mtx) );
}
/* Create a new message to pass to dispatch module */
static void dispatch_new_message(
		int appid, 
		dict_object_t * cmd, 
		dict_object_t * avp1,
		int val1, 
		dict_object_t * avp2, /* or NULL */
		int val2 )
{
	msg_t * msg = NULL;
	msg_avp_t * avp = NULL;
	msg_data_t * msgdata = NULL;
	avp_value_t value = { .i32 = val1 };
	
	/* Create the message */
	CHECK( 0, msg_new ( cmd, MSGFL_ALLOW_ETEID, &msg ) );
	CHECK( 0, msg_data ( msg, &msgdata ) );	
	msgdata->msg_appl = appid;
	
	/* Add AVP1 */
	CHECK( 0, msg_avp_new ( avp1, 0, &avp ) );
	CHECK( 0, msg_avp_setvalue ( avp, &value ) );
	CHECK( 0, msg_avp_add ( msg, MSG_BRW_FIRST_CHILD, avp ) );
	
	/* Add AVP2 */
	if (avp2) {
		CHECK( 0, msg_avp_new ( avp2, 0, &avp ) );
		value.i32 = val2;
		CHECK( 0, msg_avp_setvalue ( avp, &value ) );
		CHECK( 0, msg_avp_add ( msg, MSG_BRW_LAST_CHILD, avp ) );
	}
	
	/* If message is an answer, create also the matching request */
	if (cmd == cmd1a) {
		msg_t * qry = NULL;
		
		CHECK( 0, msg_new ( cmd1r, 0, &qry ) );
		
		CHECK( 0, msg_send( &qry, cbans, NULL) );
		
		/* Now retrieve the message in the outgoing queue, the rt_out thread is not running */
		CHECK( 0, meq_get( g_meq_outgoing, &qry ) );
		
		/* Associate with the message */
		CHECK( 0, msg_answ_associate( msg, qry ) );
	}
	
	/* reset cb_called */
	cb_called = 0;
	
	/* Push to the queue for dispatch module */
	CHECK( 0, meq_post( g_meq_local, msg ) );
	
	return;
}

static void wait_for_val( uint32_t yes, uint32_t no )
{
	/* reset the timer */
	alarm(2);
	
	CHECK( 0, pthread_mutex_lock(&test_mtx) );

	do {
		/* Check if any cb from the "no" mask was called */
		CHECK( 0, cb_called & no );

		/* Check if all cb from "yes" mask were called */
		if ( (yes & cb_called) == yes )
			break;
		
		CHECK( 0, pthread_cond_wait( &test_cnd, &test_mtx ) );
	} while (1);
	
	CHECK( 0, pthread_mutex_unlock(&test_mtx) );
}


/* Main test routine */
int main(int argc, char *argv[])
{
	disp_cb_hdl_t * hdl1 = NULL;
	disp_cb_hdl_t * hdl2 = NULL;
	disp_cb_hdl_t * hdl3 = NULL;
	
	dict_object_t * vendor = NULL;
	dict_object_t * app1 = NULL;
	dict_object_t * app2 = NULL;
	dict_object_t * app3 = NULL;
	dict_object_t * cmd2 = NULL;
	dict_object_t * avp1 = NULL;
	dict_object_t * avp2 = NULL;
	dict_object_t * enu1 = NULL;
	dict_object_t * enu2 = NULL;
	
	
	/* First, initialize the daemon modules */
	INIT_WAAAD();
	CONFIG_WAAAD();
	
	/* Kill routing module threads, so they don't interfere */
	CHECK( 0, rt_fini() );
	
	/* Create the dictionary objects */
	{
		{
			dict_vendor_data_t vendor_data    = { T_VID,  "Vendor for test" };
			CHECK( 0, dict_new ( DICT_VENDOR, &vendor_data , NULL, &vendor ) );
		}
		{
			dict_application_data_t app_data1 = { T_AID1, "Application 1 for test" };
			dict_application_data_t app_data2 = { T_AID2, "Application 2 for test" };
			dict_application_data_t app_data3 = { T_AID3, "Application 3 for test" };
			CHECK( 0, dict_new ( DICT_APPLICATION, &app_data1 , vendor, &app1 ) );
			CHECK( 0, dict_new ( DICT_APPLICATION, &app_data2 , NULL,   &app2 ) );
			CHECK( 0, dict_new ( DICT_APPLICATION, &app_data3 , NULL,   &app3 ) );
		}
		{
			dict_cmd_data_t	cmd_data1_r = { 
				T_CID1, 
				"Command 1R for test", 
				CMD_FLAG_PROXIABLE | CMD_FLAG_REQUEST, 
				CMD_FLAG_PROXIABLE | CMD_FLAG_REQUEST 
				};
			dict_cmd_data_t	cmd_data1_a = { 
				T_CID1, 
				"Command 1A for test", 
				CMD_FLAG_PROXIABLE | CMD_FLAG_REQUEST, 
				CMD_FLAG_PROXIABLE 
				};
			dict_cmd_data_t	cmd_data2 = { 
				T_CID2, 
				"Command 2 for test", 
				CMD_FLAG_PROXIABLE | CMD_FLAG_REQUEST, 
				CMD_FLAG_PROXIABLE | CMD_FLAG_REQUEST
				};
			CHECK( 0, dict_new ( DICT_COMMAND, &cmd_data1_r , NULL,   &cmd1r ) );
			CHECK( 0, dict_new ( DICT_COMMAND, &cmd_data1_a , NULL,   &cmd1a ) );
			CHECK( 0, dict_new ( DICT_COMMAND, &cmd_data2   , NULL,   &cmd2 ) );
		}
		{
			dict_object_t * type = NULL;
			dict_type_data_t tdata = {
				AVP_TYPE_INTEGER32,
				"Type for test",
				NULL,
				NULL
			};
			dict_type_enum_data_t te1data = {
				"Enum val 1 for test",
				{ .i32 = T_EV1 }
				};
			dict_type_enum_data_t te2data = {
				"Enum val 2 for test",
				{ .i32 = T_EV2 }
				};
			dict_avp_data_t avp1_data = {
				T_AVID1,
				T_VID,
				"AVP 1 for test",
				AVP_FLAG_VENDOR,
				AVP_FLAG_VENDOR,
				AVP_TYPE_INTEGER32
				};
			dict_avp_data_t avp2_data = {
				T_AVID2,
				0,
				"AVP 2 for test",
				0,
				0,
				AVP_TYPE_INTEGER32
				};
			CHECK( 0, dict_new ( DICT_TYPE, &tdata , app1, &type ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &te1data , type, &enu1 ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &te2data , type, &enu2 ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp1_data , type, &avp1 ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp2_data , type, &avp2 ) );
		}
	}
	
	/* Test registration of callbacks */
	{
		disp_cb_hdl_t * hdl = NULL;
		disp_reg_val_t  regdata;
		
		CHECK( 0, disp_register( disp_cb0, DISP_REG_ANY, NULL, &hdl ) );
		CHECK( 0, disp_unregister( hdl ) );
		CHECK( 0, disp_register( disp_cb0, DISP_REG_ANY, NULL, &hdl ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.flags = 0x3;
		regdata.app_id = app1;
		CHECK( 0, disp_register( disp_cb1, DISP_REG_APPID, &regdata, &hdl1 ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.flags = 0x1;
		regdata.app_id = app2;
		CHECK( 0, disp_register( disp_cb2, DISP_REG_APPID, &regdata, &hdl2 ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.flags = 0x3;
		regdata.app_id = app3;
		CHECK( 0, disp_register( disp_cb3, DISP_REG_APPID, &regdata, &hdl3 ) );
		
		/* Now check the configuration of waaad is created as expected */
		CHECK( 5, 		g_conf->supported_apps_nb );
		CHECK( T_AID1, 		g_conf->supported_apps_list[0].a );
		CHECK( T_VID,  		g_conf->supported_apps_list[0].v );
		CHECK( APP_TYPE_AUTH, 	g_conf->supported_apps_list[0].t );
		CHECK( T_AID1, 		g_conf->supported_apps_list[1].a );
		CHECK( T_VID,  		g_conf->supported_apps_list[1].v );
		CHECK( APP_TYPE_ACCT, 	g_conf->supported_apps_list[1].t );
		CHECK( T_AID3, 		g_conf->supported_apps_list[2].a );
		CHECK( 0,  		g_conf->supported_apps_list[2].v );
		CHECK( APP_TYPE_AUTH, 	g_conf->supported_apps_list[2].t );
		CHECK( T_AID3, 		g_conf->supported_apps_list[3].a );
		CHECK( 0,  		g_conf->supported_apps_list[3].v );
		CHECK( APP_TYPE_ACCT, 	g_conf->supported_apps_list[3].t );
		CHECK( T_AID2, 		g_conf->supported_apps_list[4].a );
		CHECK( 0,  		g_conf->supported_apps_list[4].v );
		CHECK( APP_TYPE_AUTH, 	g_conf->supported_apps_list[4].t );
		
		/* Register other callbacks with different criteria */
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.command = cmd2;
		CHECK( 0, disp_register( disp_cb4, DISP_REG_CC, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.app_id = app2;
		regdata.command = cmd1a;
		CHECK( 0, disp_register( disp_cb5, DISP_REG_CC, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.app_id = app2;
		regdata.command = cmd1r;
		CHECK( 0, disp_register( disp_cb6, DISP_REG_CC, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.app_id = app2;
		regdata.command = cmd1r;
		regdata.avp = avp1;
		CHECK( 0, disp_register( disp_cb7, DISP_REG_AVP, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.command = cmd1a;
		regdata.avp = avp1;
		regdata.value = enu1;
		CHECK( 0, disp_register( disp_cb8, DISP_REG_AVP_ENUMVAL, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.value = enu2;
		CHECK( EINVAL, disp_register( disp_cb9, DISP_REG_AVP_ENUMVAL, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.avp = avp1;
		regdata.value = enu2;
		CHECK( 0, disp_register( disp_cb9, DISP_REG_AVP_ENUMVAL, &regdata, NULL ) );
		
		memset( &regdata, 0, sizeof(disp_reg_val_t) );
		regdata.app_id = app3;
		regdata.command = cmd1a;
		regdata.avp = avp2;
		regdata.value = enu2;
		CHECK( 0, disp_register( disp_cb10, DISP_REG_AVP_ENUMVAL, &regdata, NULL ) );
		
	}
	
	{
		struct {
			int appid;
			dict_object_t * cmd;
			dict_object_t * avp1;
			int val1;
			dict_object_t * avp2;
			int val2;
			int cb[11]; /* 0: may, 1: must, -1: must not */
			int cba;
		} test_matrix[] = {
		/*	appid	cmd	avp1	val1	avp2	val2	cb0,cb1,cb2,cb3,cb4,cb5,cb6,cb7,cb8,cb9,cb10 	cba*/
		 {	0,	cmd1r,	avp1,	0,	NULL,	0,	{ 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },	 0 }
		,{	0,	cmd1r,	avp1,	T_EV1,	avp1,	0,	{ 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },	 0 }
		,{	0,	cmd1r,	avp1,	T_EV1,	avp1,	T_EV2,	{ 1, -1, -1, -1, -1, -1, -1, -1, -1,  1, -1 },	 0 }
		,{	0,	cmd1r,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1, -1, -1, -1, -1, -1, -1, -1,  1, -1 },	 0 }
		,{	0,	cmd1a,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1, -1, -1, -1, -1, -1, -1,  1,  1, -1 },	 1 }
		,{	0,	cmd2,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1, -1, -1,  1, -1, -1, -1, -1,  1, -1 },	 0 }
		,{	T_AID1,	cmd2,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1,  1, -1, -1,  1, -1, -1, -1, -1,  1, -1 },	 0 }
		,{	T_AID2,	cmd2,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1,  1, -1,  1, -1, -1, -1, -1,  1, -1 },	 0 }
		,{	T_AID2,	cmd1a,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1,  1, -1, -1,  1, -1, -1,  1,  1, -1 },	 1 }
		,{	T_AID2,	cmd1r,	avp1,	T_EV2,	avp1,	T_EV1,	{ 1, -1,  1, -1, -1, -1,  1,  1, -1,  1, -1 },	 0 }
		,{	T_AID3,	cmd1a,	avp2,	T_EV2,	avp1,	T_EV2,	{ 1, -1, -1, -1, -1, -1, -1, -1, -1,  0,  1 },	 1 }
		};
		int i;
		
		for (i=0; i< sizeof(test_matrix) / sizeof(test_matrix[0]); i++) {
			uint32_t mask_yes = 0;
			uint32_t mask_no = 0;
			int j;
		
			for (j=0; j < 11; j++) {
				if (test_matrix[i].cb[j]) {
					if (test_matrix[i].cb[j] == 1) {
						mask_yes |= 1 << j;
					} else {
						mask_no |= 1 << j;
					}
				}
			}
			if (test_matrix[i].cba)
				mask_yes |= 1 << 31;
			else
				mask_no |= 1 << 31;
			
			#if 0
			fprintf(stderr,
					"%s:%-4d: Starting next test, line %d, yes: %x, no: %x\n",
					__FILE__,
					__LINE__,
					i, mask_yes, mask_no);
			#endif
			
			dispatch_new_message(
					test_matrix[i].appid,
					test_matrix[i].cmd,
					test_matrix[i].avp1,
					test_matrix[i].val1,
					test_matrix[i].avp2,
					test_matrix[i].val2
					);
			
			wait_for_val( mask_yes, mask_no );
			
		}
	}
	
	/* Test de-registration */
	{
		CHECK( 0, disp_unregister( hdl1 ) );
		CHECK( 0, disp_unregister( hdl2 ) );
		CHECK( 0, disp_unregister( hdl3 ) );
	}
	
	/* That's all for the tests yet */
	PASSTEST();
} 
	
